VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CalcTree"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private vExpression As String, vOperator As String, vInvertSign As Boolean
Private fError As Boolean, sError As String
Private colOperands As Collection
Private objParent As CalcTree

' Operators:
'   () (parenthesis)    : (2)
'   - (invert sign, ~)  : -3, -(-A)
'   ^ (power)           : 2^3, A^2
'   %, \                : 5%3, 3\2
'   *, /                : 1*2, 3/4
'   +, -                : 1+2, 5-3, 1+4-2
'   functions           : sin(30), log(8,2)

Public Event RequestData(Field As String, Value As Variant)
Public Event RequestFunction(Name As String, Result As Variant, Arguments() As Variant)

Public Property Let Expression(ByVal NewValue As String)
Dim obj As CalcTree
vExpression = NewValue
For Each obj In colOperands
  colOperands.Remove CStr(ObjPtr(obj))
  obj.Dispose
Next obj
If vOperator <> ":" Then Parse
End Property

Public Property Get Expression() As String
Expression = vExpression
End Property

Friend Property Let Operator(ByVal NewValue As String)
vOperator = NewValue
End Property

Friend Property Get Operator() As String
Operator = vOperator
End Property

Friend Property Let InvertSign(ByVal NewValue As Boolean)
vInvertSign = NewValue
End Property

Friend Property Get InvertSign() As Boolean
InvertSign = vInvertSign
End Property

Friend Property Get Operands() As Collection
Set Operands = colOperands
End Property

Public Property Get Result() As Variant
If fError Then Exit Property
Dim obj As CalcTree, res As Variant, S As String, V As Variant, A() As Variant, I As Long
On Error GoTo ErrorHandler
If vOperator = ":" Then
  ReDim A(0 To colOperands.Count)
  For I = 1 To colOperands.Count
    Set obj = colOperands.Item(I)
    A(I) = obj.Result
    Set obj = Nothing
  Next I
  CallRequestFunction vExpression, V, A()
  If vInvertSign Then
    Result = -V
  Else
    Result = V
  End If
  Exit Property
End If
If colOperands.Count = 0 Then
  If IsNumeric(vExpression) Then
    V = Val(vExpression)
  Else
    CallRequestData vExpression, V
  End If
  If vInvertSign Then
    Result = -V
  Else
    Result = V
  End If
  Exit Property
End If
If colOperands.Count = 1 Then
  Set obj = colOperands.Item(1)
  If obj.Operator = ":" Then
    If vInvertSign Then
      Result = -obj.Result
    Else
      Result = obj.Result
    End If
    Exit Property
  End If
  Set obj = Nothing
End If
For Each obj In colOperands
  If obj.ErrorState Then
    fError = obj.Error
    sError = obj.Error
    Exit For
  End If
  V = obj.Result
  Select Case obj.Operator
    Case vbNullString
      res = V
    Case "+"
      res = res + V
    Case "-"
      res = res - V
    Case "*"
      res = res * V
    Case "/"
      res = res / V
    Case "%"
      res = res Mod V
    Case "\"
      res = res \ V
    Case "^"
      res = res ^ V
  End Select
  If fError Then Exit For
Next obj
If Not fError Then
  If vInvertSign Then
    Result = -res
  Else
    Result = res
  End If
End If
Exit Property

ErrorHandler:
fError = True
sError = Err.Description
End Property

Public Property Get ErrorState() As Boolean
ErrorState = fError
End Property

Public Property Get Error() As String
Error = sError
End Property

Private Function PairPos(Expression As String, Pos As Long) As Long
Dim I0 As Long, I As Long, I1 As Long, I2 As Long, L As Long, fReverse As Boolean
Select Case Mid$(Expression, Pos, 1)
  Case "("
    fReverse = False
  Case ")"
    fReverse = True
  Case Else
    Exit Function
End Select
I0 = Pos
If fReverse Then
  Do While I1 >= 1
    I1 = InStrRev(Expression, "(", I0)
    I2 = InStrRev(Expression, ")", I0)
    If (I1 = 0) And (I2 = 0) Then
      I = 0
    ElseIf (I1 > 0) And (I2 = 0) Then
      I = I1
    ElseIf (I1 = 0) And (I2 > 0) Then
      I = I2
    Else
      I = IIf(I1 > I2, I1, I2)
    End If
    If I = 0 Then Exit Do
    Select Case Mid$(Expression, I, 1)
      Case ")"
        L = L + 1
      Case "("
        L = L - 1
        If L = 0 Then Exit Do
    End Select
    I0 = I - 1
  Loop
Else
  Do While I0 <= Len(Expression)
    I1 = InStr(I0, Expression, "(")
    I2 = InStr(I0, Expression, ")")
    If (I1 = 0) And (I2 = 0) Then
      I = 0
    ElseIf (I1 > 0) And (I2 = 0) Then
      I = I1
    ElseIf (I1 = 0) And (I2 > 0) Then
      I = I2
    Else
      I = IIf(I1 < I2, I1, I2)
    End If
    If I = 0 Then Exit Do
    Select Case Mid$(Expression, I, 1)
      Case "("
        L = L + 1
      Case ")"
        L = L - 1
        If L = 0 Then Exit Do
    End Select
    I0 = I + 1
  Loop
End If
PairPos = I
End Function

Private Function FirstInStr(Start As Long, Text As String, ParamArray Mathes() As Variant) As Long
Dim I As Long, P As Long, ret As Long
For I = LBound(Mathes) To UBound(Mathes)
  If Len(Mathes(I)) > 0 Then
    P = InStr(Start, Text, Mathes(I))
    If P > 0 Then
      If P < ret Or ret = 0 Then ret = P
    End If
  End If
Next I
FirstInStr = ret
End Function

Private Function FirstInStrRev(Start As Long, Text As String, ParamArray Mathes() As Variant) As Long
Dim I As Long, P As Long, ret As Long
For I = LBound(Mathes) To UBound(Mathes)
  If Len(Mathes(I)) > 0 Then
    P = InStrRev(Text, Mathes(I), Start)
    If P > 0 Then
      If P > ret Or ret = 0 Then ret = P
    End If
  End If
Next I
FirstInStrRev = ret
End Function

Private Function CompactParenths(Expression As String, Parts As Collection) As String
Dim I0 As Long, I As Long, P As Long, S As String, res As String, fFunction As Boolean
Set Parts = New Collection
I0 = 1
Do Until I0 > Len(Expression)
  I = InStr(I0, Expression, "(")
  If I = 0 Then
    I = Len(Expression) + 1
    S = Mid$(Expression, I0, I - I0)
  Else
    fFunction = False
    If I > I0 Then
      If InStr("()-^%\*/+-", Mid$(Expression, I - 1, 1)) = 0 Then
        fFunction = True
      End If
    End If
    If fFunction Then
      P = PairPos(Expression, I)
      If P = 0 Then
        fError = True
        sError = "Unclosed parenthesis"
        Exit Function
      Else
        I = FirstInStrRev(I - 1, Expression, ",", "(", ")", "^", "%", "\", "*", "/", "+", "-") + 1
        res = res & Mid$(Expression, I0, I - I0)
        Parts.Add Mid$(Expression, I, P - I + 1)
        S = "#" & Trim$(Str$(Parts.Count)) & "#"
        I = P
      End If
    Else
      P = PairPos(Expression, I)
      If P = 0 Then
        fError = True
        sError = "Unclosed parenthesis"
        Exit Function
      Else
        res = res & Mid$(Expression, I0, I - I0)
        S = Mid$(Expression, I, P - I + 1)
        Parts.Add S
        S = "#" & Trim$(Str$(Parts.Count)) & "#"
        I = P
      End If
    End If
  End If
  res = res & S
  I0 = I + 1
Loop
CompactParenths = res
End Function

Private Function ExpandParenths(Expression As String, Parts As Collection) As String
If Parts Is Nothing Then
  ExpandParenths = Expression
  Exit Function
End If
Dim I0 As Long, I As Long, res As String
I0 = 1
Do Until I0 > Len(Expression)
  I = InStr(I0, Expression, "#")
  If I = 0 Then
    res = res & Mid$(Expression, I0)
    Exit Do
  Else
    res = res & Mid$(Expression, I0, I - I0)
    I0 = I + 1
    I = InStr(I0, Expression, "#")
    If I = 0 Then Exit Do
    res = res & Parts.Item(Val(Mid$(Expression, I0, I - I0)))
  End If
  I0 = I + 1
Loop
ExpandParenths = res
End Function

Private Sub Parse()
Dim obj As CalcTree, Parts As Collection, V As Variant, E As String, P As Long
Dim subcol As Collection, subobj As CalcTree, subparts As Collection, SI0 As Long, SI As Long
Dim I0 As Long, I As Long, S As String, Op As String, O1 As String, O2 As String, fMinus As Boolean
fError = False
sError = vbNullString
E = CompactParenths(vExpression, Parts)
If fError Then Exit Sub
I0 = 1
Do Until I0 > Len(E)
  I = InStr(I0, E, "-")
  If I = 0 Then Exit Do
  If I = I0 Then
    Mid$(E, I, 1) = "~"
  Else
    If InStr("^%\*/+-", Mid$(E, I - 1, 1)) Then
      Mid$(E, I, 1) = "~"
    End If
  End If
  I0 = I + 1
Loop
If FirstInStr(1, E, "+", "-") > 0 Then
  O1 = "+": O2 = "-"
ElseIf FirstInStr(1, E, "*", "/") > 0 Then
  O1 = "*": O2 = "/"
ElseIf FirstInStr(1, E, "%", "\") > 0 Then
  O1 = "%": O2 = "\"
ElseIf FirstInStr(1, E, "^") > 0 Then
  O1 = "^": O2 = vbNullString
Else
  S = Replace$(ExpandParenths(E, Parts), "~", "-")
  If Left$(S, 1) = "-" Then
    fMinus = True
    S = Mid$(S, 2)
  End If
  If Left$(S, 1) = "(" And Right$(S, 1) = ")" Then
    Set obj = New CalcTree
    obj.Create Me
    obj.InvertSign = fMinus
    obj.Expression = Mid$(S, 2, Len(S) - 2)
    colOperands.Add Key:=CStr(ObjPtr(obj)), Item:=obj
    Set obj = Nothing
  ElseIf InStr(S, "(") > 0 And Right$(S, 1) = ")" Then
    E = Left$(S, InStr(S, "(") - 1)
    S = Mid$(S, InStr(S, "("))
    Set obj = New CalcTree
    obj.Create Me
    obj.InvertSign = fMinus
    obj.Operator = ":"
    obj.Expression = E
    Set subcol = obj.Operands
    E = CompactParenths(Mid$(S, 2, Len(S) - 2), subparts)
    SI0 = 1
    Do Until SI0 > Len(E)
      SI = InStr(SI0, E, ",")
      If SI = 0 Then SI = Len(E) + 1
      Set subobj = New CalcTree
      subobj.Create obj
      subobj.Expression = ExpandParenths(Mid$(E, SI0, SI - SI0), subparts)
      subcol.Add Key:=CStr(ObjPtr(subobj)), Item:=subobj
      Set subobj = Nothing
      SI0 = SI + 1
    Loop
    Set subparts = Nothing
    Set subcol = Nothing
    colOperands.Add Key:=CStr(ObjPtr(obj)), Item:=obj
    Set obj = Nothing
  Else
    vExpression = S
    vInvertSign = fMinus
  End If
  Exit Sub
End If
I0 = 1
Do Until I0 > Len(E)
  I = FirstInStr(I0, E, O1, O2)
  If I = 0 Then I = Len(E) + 1
  S = Mid$(E, I0, I - I0)
  If Len(S) > 0 Then
    S = Replace$(ExpandParenths(S, Parts), "~", "-")
    Set obj = New CalcTree
    obj.Create Me
    obj.Operator = Op
    obj.Expression = S
    colOperands.Add Key:=CStr(ObjPtr(obj)), Item:=obj
    Set obj = Nothing
  End If
  Op = Mid$(E, I, 1)
  I0 = I + 1
Loop
Set Parts = Nothing
End Sub

Friend Sub CallRequestData(Field As String, Value As Variant)
If objParent Is Nothing Then
  RaiseEvent RequestData(Field, Value)
Else
  objParent.CallRequestData Field, Value
End If
End Sub

Friend Sub CallRequestFunction(Name As String, Result As Variant, Arguments() As Variant)
If objParent Is Nothing Then
  RaiseEvent RequestFunction(Name, Result, Arguments())
Else
  objParent.CallRequestFunction Name, Result, Arguments()
End If
End Sub

Friend Sub Create(Parent As CalcTree)
Set objParent = Parent
End Sub

Friend Sub Dispose()
Dim obj As CalcTree
For Each obj In colOperands
  obj.Dispose
Next obj
Set objParent = Nothing
End Sub

Private Sub Class_Initialize()
Set colOperands = New Collection
End Sub

Private Sub Class_Terminate()
Set colOperands = Nothing
End Sub
